//*********************************************************
//
//               HCS - HeroChooseSystem v1.2f
//
//                       by Blubb-Tec
//
//*********************************************************
//=========================================================
//*********************************************************
//
// Configuration is further down.
// If you found this system in some other map, please look
// for the latest version and a better documentation under:
// [url]http://www.wc3campaigns.net/showthread.php?p=1024151[/url]
//
//**********************************************************
//==========================================================
//**********************************************************

library HCS initializer init
// dirty lil' helpers begin here -
globals
    private integer randomcounter
    private unit randomunit
    private hashtable ht
endglobals

private function GroupGetRandomUnitEnum takes nothing returns nothing
    if IsUnitHidden(GetEnumUnit()) then
        return
    else
        if randomcounter == 0 then
            set randomunit = GetEnumUnit()
        endif
        set randomcounter = randomcounter - 1
    endif
endfunction
private function GroupGetRandomUnit takes group g, integer count returns unit
    set randomunit = null
    set randomcounter = GetRandomInt(0,count-1)
    call ForGroup(g, function GroupGetRandomUnitEnum)
    return randomunit
endfunction

private function SetUnitVal takes unit u, integer val returns nothing
    call SaveInteger(ht, StringHash(SCOPE_PREFIX+"_____"), GetHandleId(u), val)
endfunction

private function GetUnitVal takes unit u returns integer
    return LoadInteger(ht, StringHash(SCOPE_PREFIX+"_____"), GetHandleId(u))
endfunction
// - and end here.

// CONFIGURATION
globals
    // DummyPlayer: the player who will own the choosable heroes,
    // and over whom all players will gain shared units.
    // make sure no other units of this player are visible during the herochoose,
    // or the players will be able to control them.
    public constant player  DummyPlayer = Player(15) //Neutral Extra
    
    // DummyID: the ID of the dummy
    public constant integer DummyID = 'e000'
    
    // SoldDummyID: the id of the dummy unit that will serve as a 'tooltip'
    // for the "pick-this-hero"-button
    public constant integer SoldDummyID = 'e001'
    // SoldRandomID: same as above, but it will only be added to the hero
    // that serves as a random-chooser, so you can add another tooltip
    // for that one.
    public constant integer SoldRandomID = 'e002'
    
    // ChooseAbilID: the ID of the HCS Choose Unit ability you copied into your map
    public constant integer ChooseAbilID = 'A000'
    // randomPickerID: the ID of the hero that will serve as a random-hero-picker
    // whenever this unit can be choosen as a hero, it will give a random hero instead
    public constant integer randomPickerID = 'N000'
    
    // multiPick: defines if a single hero can be picked multiple times
    public constant boolean multiPick = false
    
    // HCDuration: the duration for which the Herochoose is running(in seconds)
    public constant real HCDuration = 30.
    
    // randomHeroesAtEnd: defines if players without heroes shall be given random ones
    // when the hero choose is over
    public constant boolean randomHeroesAtEnd = true
    
    // ErrorMsg: the message that will be given to a player who wants to choose a hero he can't
    public constant string  ErrorMsg = "This Hero belongs to another team - you can't pick it!"
    
    // EyeCandyChoose: this string determines which effect shall be used on chosen heroes
    public constant string  EyeCandyChoose = "Abilities\\Spells\\Human\\ReviveHuman\\ReviveHuman.mdl"
    
    // EyeCandyCreate: this string determines which effect shall be used on spawned Heroes
    public constant string  EyeCandyCreate = "Abilities\\Spells\\Human\\ReviveHuman\\ReviveHuman.mdl"
endglobals
// CONFIGURATION END

// library code begins here -
struct FullAllianceSetting
    private boolean array s[8]
    private player p
    
    static method create takes player p returns FullAllianceSetting
        local FullAllianceSetting f = FullAllianceSetting.allocate()
        set f.p = p
        set f.s[0] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_PASSIVE)
        set f.s[1] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_HELP_REQUEST)
        set f.s[2] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_HELP_RESPONSE)
        set f.s[3] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_XP)
        set f.s[4] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_SPELLS)
        set f.s[5] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_CONTROL)
        set f.s[6] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_VISION)
        set f.s[7] = GetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_ADVANCED_CONTROL)
        return f
    endmethod
    
    method onDestroy takes nothing returns nothing
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_PASSIVE, .s[0])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_HELP_REQUEST, .s[1])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_HELP_RESPONSE, .s[2])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_SHARED_XP, .s[3])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_SHARED_SPELLS, .s[4])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_SHARED_CONTROL, .s[5])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_SHARED_VISION, .s[6])
        call SetPlayerAlliance(DummyPlayer, .p, ALLIANCE_SHARED_ADVANCED_CONTROL, .s[7])        
    endmethod
endstruct

public function interface CallbackFunc takes unit created returns nothing
public function interface RepickCallback takes player p returns integer
globals
    private FullAllianceSetting array Alliances
    
    private RepickCallback repickCallback
    
    private HeroChooser array AllHeroChooser
    private integer HeroChooserCount = 0
    
    private unit array dummys
    
    private trigger cancelOrders
    private trigger unitBought
    private trigger repickTrig
    
    private camerasetup array cams
    private timer PeriodicCams = CreateTimer()
    
    private timer HCTimer = CreateTimer()
    private timerdialog HCTimerDialog = CreateTimerDialog(HCTimer)
    
    private camerasetup gcam
    private integer RepickID = 0
    private force geforce
    private group ggroup
    private integer gindex
    private integer enumcounter
    private unit rndPicker = null
endglobals

struct HeroChooser
    private group heroes = CreateGroup()
    private force team = CreateForce()
    private camerasetup cam
    private CallbackFunc callback
    private unit randomPicker
    private boolean multiPick
    private boolean randomHeroes
    private integer index = -1
    private integer herocount
    private real SpawnX
    private real SpawnY
    
    static method create takes group Heroes, force Team, camerasetup Cam, CallbackFunc Callback, integer Index, real spawnX, real spawnY returns HeroChooser
        local HeroChooser a = HeroChooser.allocate()
        set a.heroes = Heroes
        set a.team = Team
        set a.cam = Cam
        set a.callback = Callback
        set a.index = Index
        set a.SpawnX = spawnX
        set a.SpawnY = spawnY
        return a
    endmethod
    
    static method InitSinglePlayer takes player p returns nothing
        local FullAllianceSetting f = FullAllianceSetting.create(p)
        local unit u = dummys[GetPlayerId(p)]
        if u == null then
            set u = CreateUnit(p, DummyID, CameraSetupGetDestPositionX(gcam), CameraSetupGetDestPositionY(gcam), 0.)
            set dummys[GetPlayerId(p)] = u
        endif        
        set Alliances[GetPlayerId(p)] = f
        set cams[GetPlayerId(p)] = gcam
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_PASSIVE,       true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_HELP_REQUEST,  true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_HELP_RESPONSE, true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_XP,     true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_SPELLS, true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_CONTROL, true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_VISION, true)
        call SetPlayerAlliance(DummyPlayer, p, ALLIANCE_SHARED_ADVANCED_CONTROL, false)        
        if GetLocalPlayer() == p then
            call SetCameraTargetController(u, 0., 0., true)
            call CameraSetupApplyForceDuration(gcam, false, 0.)
        endif
        set u = null
    endmethod
    
    static method PreparePlayersEnum takes nothing returns nothing
        call HeroChooser.InitSinglePlayer(GetEnumPlayer())
    endmethod
    
    static method PrepareHeroesEnum takes nothing returns nothing
        if (RepickID != 0) then
            if GetUnitTypeId(GetEnumUnit()) == RepickID then
                call ShowUnit(GetEnumUnit(), true)
                call PauseUnit(GetEnumUnit(), false)
            endif
            return
        endif
        call ShowUnit(GetEnumUnit(), true)
        call PauseUnit(GetEnumUnit(), false)
        if GetUnitTypeId(GetEnumUnit()) == randomPickerID then
            set rndPicker = GetEnumUnit()
        endif
        
        set enumcounter = enumcounter + 1
        if GetUnitVal(GetEnumUnit()) != 0 then
            call SetUnitVal(GetEnumUnit(), gindex)            
            return
        endif
        call SetUnitOwner(GetEnumUnit(), DummyPlayer, false)
        call SetUnitInvulnerable(GetEnumUnit(), true)
        call UnitAddAbility(GetEnumUnit(), ChooseAbilID)
        call UnitAddAbility(GetEnumUnit(), 'Asud')
        call UnitAddAbility(GetEnumUnit(), 'Abun')
        call UnitRemoveAbility(GetEnumUnit(), 'Amov')
        call SetUnitPathing(GetEnumUnit(), false)
        call SetUnitVal(GetEnumUnit(), gindex)
        if GetUnitTypeId(GetEnumUnit()) == randomPickerID then
            call AddUnitToStock(GetEnumUnit(), SoldRandomID, 1, 1)
        else
            call AddUnitToStock(GetEnumUnit(), SoldDummyID, 1, 1)
        endif
    endmethod
    
    method start takes nothing returns nothing
        set geforce = .team
        set gcam = .cam
        set gindex = .index + 1
        set enumcounter = 0
        set ggroup = .heroes
        set rndPicker = null
        call ForGroup(ggroup, function HeroChooser.PrepareHeroesEnum)
        call ForForce(geforce, function HeroChooser.PreparePlayersEnum)
        set .herocount = enumcounter
        set .randomPicker = rndPicker
        set ggroup = null
        set rndPicker = null
        set geforce = null
        set gindex = -1
    endmethod
    
    method assignRandomHero takes player whichPlayer returns nothing
        if .randomPicker != null then
            call GroupRemoveUnit(.heroes, .randomPicker)
            set .herocount = .herocount - 1
        endif
        if .herocount != 0 then
            call .picked(GroupGetRandomUnit(.heroes,.herocount),whichPlayer)
        endif
        if .randomPicker != null then                
            call GroupAddUnit(.heroes, .randomPicker)
            set .herocount = .herocount + 1
        endif
    endmethod
    
    method picked takes unit whichHero, player whichPlayer returns nothing
        local integer id = GetUnitTypeId(whichHero)
        local unit u

        if id == randomPickerID then
            call .assignRandomHero(whichPlayer)
            return
        endif
        //call RemoveUnit(dummys[GetPlayerId(whichPlayer)])
        //set dummys[GetPlayerId(whichPlayer)] = null
        call DestroyEffect(AddSpecialEffect(EyeCandyChoose, GetUnitX(whichHero), GetUnitY(whichHero)))
        if not multiPick then
            call ShowUnit(whichHero, false)
            call PauseUnit(whichHero, true)
            set .herocount = .herocount - 1
        endif
        
        call Alliances[GetPlayerId(whichPlayer)].destroy()
        set Alliances[GetPlayerId(whichPlayer)] = 0
        set cams[GetPlayerId(whichPlayer)] = null
        set u = CreateUnit(whichPlayer, id, .SpawnX, .SpawnY, 0.)
        call DestroyEffect(AddSpecialEffectTarget(EyeCandyCreate, u, "origin"))
        if GetLocalPlayer() == whichPlayer then
            call ResetToGameCamera(0.)
            call PanCameraToTimed(.SpawnX, .SpawnY, 0.)
            call ClearSelection()
            call SelectUnit(u, true)
        endif
        call .callback.evaluate(u)
        return
    endmethod
    
    method check takes player p returns boolean
        return Alliances[GetPlayerId(p)] != 0
    endmethod
    
    method checkBasic takes player p returns boolean
        return IsPlayerInForce(p, .team)
    endmethod
    
    method Repick takes player p, integer unitid returns nothing
        set gcam = .cam
        call HeroChooser.InitSinglePlayer(p)
        set gcam = null
        if not multiPick then
            set .herocount = .herocount +1 
            set RepickID = unitid
            call ForGroup(.heroes, function HeroChooser.PrepareHeroesEnum)
            set RepickID = 0
        endif
    endmethod
    
    static method CleanupPlayers takes nothing returns nothing
        if not HeroChooser(gindex).check(GetEnumPlayer()) then
            call RemoveUnit(dummys[GetPlayerId(GetEnumPlayer())])
            set dummys[GetPlayerId(GetEnumPlayer())] = null
            return
        endif
        if randomHeroesAtEnd then
            call HeroChooser(gindex).assignRandomHero(GetEnumPlayer())
        else
            call Alliances[GetPlayerId(GetEnumPlayer())].destroy()
            set Alliances[GetPlayerId(GetEnumPlayer())] = 0
            set cams[GetPlayerId(GetEnumPlayer())] = null
            if GetLocalPlayer() == GetEnumPlayer() then
                call ResetToGameCamera(0.)
            endif
            call RemoveUnit(dummys[GetPlayerId(GetEnumPlayer())])
            set dummys[GetPlayerId(GetEnumPlayer())] = null
        endif
    endmethod
    
    static method HideHeroesEnum takes nothing returns nothing
        call ShowUnit(GetEnumUnit(), false)
        call PauseUnit(GetEnumUnit(), true)
    endmethod
    
    method onDestroy takes nothing returns nothing
        set gindex = integer(this)
        call ForForce(.team, function HeroChooser.CleanupPlayers)
        call ForGroup(.heroes, function HeroChooser.HideHeroesEnum)
        
        set .heroes = null
        set .team = null
    endmethod
endstruct

private function PeriodicCamReset takes nothing returns nothing
    local integer i = 0
    loop
        exitwhen i >= 16
        if (GetLocalPlayer() == Player(i)) and (cams[i] != null) then
            call SetCameraTargetController(dummys[i], 0., 0., true)
            call CameraSetupApplyForceDuration(cams[i], false, 0.)
        endif
        set i = i + 1
    endloop
endfunction

public function AddTeam takes group heroes, force team, camerasetup cam, CallbackFunc callback, real SpawnX, real SpawnY returns nothing
    local HeroChooser hc = HeroChooser.create(heroes, team, cam, callback, HeroChooserCount, SpawnX, SpawnY)
    set AllHeroChooser[HeroChooserCount] = hc
    set HeroChooserCount = HeroChooserCount + 1
endfunction

public function EnableRepick takes RepickCallback rc returns nothing
    call EnableTrigger(repickTrig)
    set repickCallback = rc
endfunction

public function End takes nothing returns nothing
    local integer i = 0
    call DisableTrigger(unitBought)
    call DisableTrigger(cancelOrders)
    call DisableTrigger(repickTrig)
    set repickCallback = 0
    call PauseTimer(PeriodicCams)
    loop
        exitwhen HeroChooserCount <= 0
        set HeroChooserCount = HeroChooserCount - 1
        call AllHeroChooser[HeroChooserCount].destroy()
    endloop

    loop
        exitwhen i >= bj_MAX_PLAYERS
        if dummys[i] != null then
            call RemoveUnit(dummys[i])
            set dummys[i] = null
        endif
        if Alliances[i] != 0 then
            call Alliances[i].destroy()
            set Alliances[i] = 0
        endif            
        set i = i + 1
    endloop
    
    call PauseTimer(HCTimer)
    call TimerDialogDisplay(HCTimerDialog, false)
endfunction

private function DelayedTimer takes nothing returns nothing
    call TimerStart(HCTimer, HCDuration, false, function End)
    call TimerDialogSetTitle(HCTimerDialog, "Herochoose")
    call TimerDialogDisplay(HCTimerDialog, true)
endfunction
public function Start takes nothing returns nothing
    local integer i = 0
    
    call EnableTrigger(unitBought)
    call EnableTrigger(cancelOrders)
    
    loop
        exitwhen i >= HeroChooserCount
        
            call AllHeroChooser[i].start()
            
        set i = i + 1
    endloop
    
    call TimerStart(PeriodicCams, 0.04, true, function PeriodicCamReset)
    call TimerStart(HCTimer, 0.1, false, function DelayedTimer)
endfunction

public function RepickHero takes integer repickIndex, integer unitid, player owner returns nothing
    if not AllHeroChooser[repickIndex-1].check(owner) then
        call AllHeroChooser[repickIndex-1].Repick(owner, unitid)
    endif
endfunction

public function IsRunning takes nothing returns boolean
    return TimerGetRemaining(HCTimer) > 0.
endfunction

private function RepickActions takes nothing returns boolean
    local integer i = 0
    local player p = GetTriggerPlayer()
    
    if not IsRunning() then
        return false
    endif
    
    loop
        exitwhen i >= HeroChooserCount
            if AllHeroChooser[i].checkBasic(p) then
                if not AllHeroChooser[i].check(p) then
                    call AllHeroChooser[i].Repick(p, repickCallback.evaluate(p))
                endif
                return false
            endif
        set i = i + 1
    endloop
    
    return false
endfunction

private function cancelOrdersConditions takes nothing returns boolean
    if GetUnitAbilityLevel(GetOrderedUnit(), ChooseAbilID) == 0 then
        return false
    endif
    if (GetIssuedOrderId() == SoldDummyID) or (GetIssuedOrderId() == SoldRandomID) then
        return false
    endif
    if OrderId("stop") == GetIssuedOrderId() then
        return false
    endif
    return true
endfunction

private function cancelOrdersActions takes nothing returns nothing
    call DisableTrigger(cancelOrders)
    call PauseUnit(GetOrderedUnit(), true)
    call IssueImmediateOrder(GetOrderedUnit(), "stop")
    call PauseUnit(GetOrderedUnit(), false)
    call EnableTrigger(cancelOrders)
endfunction

private function unitBoughtConditions takes nothing returns boolean
    if ( (GetUnitTypeId(GetSoldUnit()) == SoldDummyID) or (GetUnitTypeId(GetSoldUnit()) == SoldRandomID) ) and (GetUnitTypeId(GetBuyingUnit()) == DummyID) then
        if not AllHeroChooser[GetUnitVal(GetSellingUnit())-1].check(GetOwningPlayer(GetSoldUnit())) then
            call DisplayTextToPlayer( GetOwningPlayer(GetSoldUnit()), 0.80, 0, "|c00FFFF00" + ErrorMsg + "|r" )
            call RemoveUnit(GetSoldUnit())
            return false
        endif
        return true
    endif
    
    call RemoveUnit(GetSoldUnit())
    return false
endfunction

private function unitBoughtActions takes nothing returns nothing
    call AllHeroChooser[GetUnitVal(GetSellingUnit())-1].picked(GetSellingUnit(),GetOwningPlayer(GetSoldUnit()))
    call RemoveUnit(GetSoldUnit())
endfunction

private function init takes nothing returns nothing
    local integer i = 0
    set cancelOrders = CreateTrigger()
    set unitBought = CreateTrigger()
    set repickTrig = CreateTrigger()
    set ht = InitHashtable()
    
    loop
        exitwhen i >= 15
            set Alliances[i] = 0
            set dummys[i] = null
            call TriggerRegisterPlayerChatEvent(repickTrig, Player(i), "-repick", true)
        set i = i + 1
    endloop
    
    call TriggerRegisterPlayerUnitEvent(cancelOrders,DummyPlayer,EVENT_PLAYER_UNIT_ISSUED_TARGET_ORDER,null)
    call TriggerRegisterPlayerUnitEvent(cancelOrders,DummyPlayer,EVENT_PLAYER_UNIT_ISSUED_POINT_ORDER,null)
    call TriggerRegisterPlayerUnitEvent(cancelOrders,DummyPlayer,EVENT_PLAYER_UNIT_ISSUED_ORDER,null)
    call TriggerRegisterPlayerUnitEvent(unitBought,DummyPlayer,EVENT_PLAYER_UNIT_SELL,null)
    
    call TriggerAddAction(unitBought, function unitBoughtActions)
    call TriggerAddCondition(cancelOrders, Condition(function cancelOrdersConditions))
    call TriggerAddAction(cancelOrders, function cancelOrdersActions)
    call TriggerAddCondition(unitBought, Condition(function unitBoughtConditions))
    call TriggerAddCondition(repickTrig, Condition(function RepickActions))
    
    call DisableTrigger(unitBought)
    call DisableTrigger(cancelOrders)
endfunction
// - and ends here.
endlibrary

// Thanks for using :)